All irccdctl commands
=====================

This file groups all irccdctl commands, help messages so that it's easier to maintain the manual pages, HTML
documentation and command line help in case of change.

This file can also been reviewed in case of spoken errors.

Please update this file before updating any source code.

All values that are marked optional MUST be silently discarded in the commands if they are not valid.

# JavaScript API

Whenever possible, if a function in the irccd's API is available, the arguments MUST be defined in the same order. For
example, the method `Server.cnotice` should have **channel**, **message** arguments.

# Updating the code

Adding a new transport command is quite straightforward.

You need to create two classes: one as irccdctl request and one as irccd command. Both files must be named
`command-<name>.(h|cpp)`.

# Irccd command

To implement the command irccd's side, you must write a class like this:

**command-mycommand.h**

Inherit from `TransportCommand` and override the `exec` function.

    #include "transport-command.h"

    namespace irccd {

    namespace command {

    class MyCommand : public TransportCommand {
    public:
	void exec(Irccd &irccd, TransportClient &tc, const json::Value &object) const override
	{
		// Do what you want here
	}
    };

    } // !command

    } // !irccd

Then, in the `irccd` main.cpp file, register the new class using `Irccd::addTransportCommand` function.

# Irccdctl command

In irccdctl, you must implement two functions. One to show the usage when calling `irccdctl help` and the one to
execute.

Inherit from `Command` and override both `usage` and `exec` functions.

    #include "command.h"

    namespace irccd {

    namespace command {

    class MyCommand : public Command {
    public:
	void usage(Irccdctl &irccdctl) const
	{
		// print the message using log::warning().
	}

	void exec(Irccdctl &irccdctl, const std::vector<std::string> &args) const
	{
		// send the command using irccdctl.send().
	}
    };

    } // !command

    } // !irccd

And like irccd, you must register it in the `irccdctl` main.cpp file.

# Updating the documentation

  1. You must update doc/html/guide/05-irccdctl/02-commands.md to show the command usage,
  2. You must update doc/man/irccdctl.1.in to show the command usage too,
  3. You must update the appropriate command class in both irccd/irccdctl.

**Note:** you must use the text from this file and propagate it to the files described above.

# Server commands

## server-cmode

Change the mode of the specified channel.

### Usage

`irccdctl server-cmode server channel mode`

### Arguments

  - **command**: (string) "server-cmode",
  - **server**: (string) the server unique id,
  - **channel**: (string) the channel,
  - **mode**: (string) the argument.

### Responses

  - An error if the server does not exist.

## server-cnotice

Send a message notice on a channel.

### Usage

`server-cnotice server channel message`

### Arguments

  - **command**: (string) "server-cnotice",
  - **server**: (string) the server unique id,
  - **channel**: (string) the channel name,
  - **message**: (string) the notice message.

## server-connect

Connect to a server.

### Usage

`server-connect [options] id host [port]`

**Options:**

  - **-c, --command**: specify the command char,
  - **-n, --nickname**: specify a nickname,
  - **-r, --realname**: specify a real name,
  - **-u, --username**: specify a user name,
  - **-S, --ssl-verify**: verify SSL,
  - **-s, --ssl**: connect using SSL,

### Arguments

  - **command**: (string) "server-connect",
  - **name**: (string) the server unique id,
  - **host**: (string) the host address,
  - **port**: (int) the port number (Optional, default: 6667),
  - **ssl**: (bool) use SSL (Optional, default: false),
  - **sslVerify**: (bool) verify SSL (Optional, default: false),
  - **nickname**: (string) the nickname to use (Optional, default: irccd),
  - **username**: (string) the user name to use (Optional, default: irccd),
  - **realname**: (string) the real name to use (Optional, default: IRC Client Daemon),
  - **ctcpVersion**: (string) the CTCP Version to answer (Optional, default: the irccd's version),
  - **commandChar**: (string) the command character to use to invoke commands (Optional, default: !),
  - **reconnectTries**: (int) the number of reconnection to try (Optional, default: -1),
  - **reconnectTimeout**: (int) the number of seconds to wait before retrying to connect (Optional, default: 30).

### Responses

  - An error if a server with the same name exists.

## server-disconnect

Disconnect from a server.

If server is not specified, irccd disconnects all servers.

### Usage

`server-disconnect [name]`

### Arguments

  - **command**: (string) "server-disconnect",
  - **server**: (string) the server unique id (Optional, default: none).

### Responses

  - An error if the server does not exist.

## server-info

Get server information.

### Usage

`server-info name`

### Arguments

  - **command**: (string) "server-info",
  - **server**: (string) the server unique id.

### Responses

  - The following properties on success:
    - **name**: (string) the server unique id,
    - **host**: (string) the server hostname,
    - **port**: (int) the port,
    - **ipv6**: (bool) true if using IPv6,
    - **ssl**: (bool) true if connection is using SSL,
    - **sslVerify**: (bool) true if SSL was verified,
    - **channels**: (string list) list of channels.
    - **nickname**: (string) the current nickname in use,
    - **username**: (string) the username in use,
    - **realname**: (string) the realname in use.
  - An error if the server does not exist.

### Responses

  - An error if the server does not exist.

## server-invite

Invite the specified target on the channel.

### Usage

`server-invite server target channel`

### Arguments

  - **command**: (string) "server-invite",
  - **server**: (string) the server unique id,
  - **target**: (string) the nickname to invite,
  - **channel**: (string) the channel.

## server-join

Join the specified channel, the password is optional.

### Usage

`server-join server channel [password]`

### Arguments

  - **command**: (string) "server-join",
  - **server**: (string) the server unique id,
  - **channel**: (string) the channel to join,
  - **password**: (string) the password (Optional, default: none).

### Responses

  - An error if the server does not exist.

## server-kick

Kick the specified target from the channel, the reason is optional.

### Usage

`server-kick server target channel [reason]`

### Arguments

  - **command**: (string) "server-kick",
  - **server**: (string) the server unique id,
  - **target**: (string) the target nickname,
  - **channel**: (string) the channel,
  - **reason**: (string) the reason (Optional, default: none).

### Responses

  - An error if the server does not exist.

## server-list

Get the list of all connected servers.

### Usage

`irccdctl server-list`

### Arguments

  - **command**: (string) "server-list".

### Responses

  - The following properties:
    - **list**: (string list) the list of all server unique ids.

## server-me

Send an action emote.

### Usage

`irccdctl server-me server target message`

### Arguments

  - **command**: (string) "server-me",
  - **server**: (string) the server unique id,
  - **target**: (string) the target or channel,
  - **message**: (string) the message.

### Responses

  - An error if the server does not exist.

## server-message

Send a message to the specified target or channel.

### Usage

`irccdctl server-message server target message`

### Arguments

  - **command**: (string) "server-message",
  - **server**: (string) the server unique id,
  - **target**: (string) the target or channel,
  - **message**: (string) the message.

### Responses

  - An error if the server does not exist.

## server-mode

Set the irccd's user mode.

### Usage

`irccdctl server-mode server mode`

### Arguments

  - **command**: (string) "server-mode",
  - **server**: (string) the server unique id,
  - **mode**: (string) the mode.

### Responses

  - An error if server is specified and the server does not exist.

## server-nick

Change irccd's nickname.

### Usage

`irccdctl server-nick server nickname`

### Arguments

  - **command**: (string) "server-nick",
  - **server**: (string) the server unique id,
  - **nickname**: (string) the new nickname.

### Responses

  - An error if the server does not exist.

## server-notice

Send a private notice to the specified target.

### Usage

`irccdctl server-notice server target message`

### Arguments

  - **command**: (string) "server-notice",
  - **server**: (string) the server unique id,
  - **target**: (string) the target,
  - **message**: (string) the notice message.

### Responses

  - An error if the server does not exist.

## server-part

Leave the specified channel, the reason is optional.

Not all IRC servers support giving a reason to leave a channel, do not specify it if this is a concern.

### Usage

`irccdctl server-part server channel [reason]`

### Arguments

  - **command**: (string) "server-part",
  - **server**: (string) the unique server id,
  - **channel**: (string) the channel to leave,
  - **reason**: (string) the reason (Optional, default: none).

### Responses

  - An error if the server does not exist.

## server-reconnect

Force reconnection of one or all servers.

If server is not specified, all servers will try to reconnect.

### Usage

`irccdctl server-reconnect [server]`

### Arguments

  - **command**: (string) "server-reconnect",
  - **server**: (string) the server unique id (Optional, default: none).

### Responses

  - An error if server is specified and the server does not exist.

## server-topic

Change the topic of the specified channel.

### Usage

`irccdctl server-topic server channel topic`

### Arguments

  - **command**: (string) "server-topic",
  - **server**: (string) the unique server id,
  - **channel**: (string) the channel,
  - **topic**: (string) the new topic.

### Responses

  - An error if server is specified and the server does not exist.
