/*
 * main.cpp -- test Irccd.ElapsedTimer API
 *
 * Copyright (c) 2013-2016 David Demelier <markand@malikania.fr>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <gtest/gtest.h>

#include <thread>

#include <js-irccd.h>
#include <js-elapsed-timer.h>

using namespace irccd;
using namespace std::chrono_literals;

/*
 * For all tests, we tolerate 30 ms because some systems have bigger lags.
 */
static constexpr int margin = 30;

class TestElapsedTimer : public testing::Test {
protected:
	js::Context m_context;

	TestElapsedTimer()
	{
		loadJsIrccd(m_context);
		loadJsElapsedTimer(m_context);
	}

	inline void assertRange(int value, int expected) const noexcept
	{
		if (value < (expected - margin) || value > (expected + margin))
			FAIL() << value << " is bigger than [" << (expected - margin) << ", " << (expected + margin) << "]";
	}
};

TEST_F(TestElapsedTimer, standard)
{
	m_context.peval(js::Script{"timer = new Irccd.ElapsedTimer();"});

	std::this_thread::sleep_for(300ms);

	m_context.peval(js::Script{"result = timer.elapsed();"});
	assertRange(m_context.getGlobal<int>("result"), 300);
}

TEST_F(TestElapsedTimer, reset)
{
	m_context.peval(js::Script{"timer = new Irccd.ElapsedTimer();"});

	std::this_thread::sleep_for(300ms);

	m_context.peval(js::Script{"timer.reset(); result = timer.elapsed();"});
	assertRange(m_context.getGlobal<int>("result"), 0);
}

TEST_F(TestElapsedTimer, pause)
{
	m_context.peval(js::Script{"timer = new Irccd.ElapsedTimer();"});

	/*
	 * Simulate a pause in the game like this:
	 *
	 * start     pause restart elapsed
	 * |   10ms   |.5ms.| 6ms  |
	 *
	 * Since the game was paused, the 5ms must not be totalized.
	 */
	std::this_thread::sleep_for(10ms);

	m_context.peval(js::Script{"timer.pause();"});

	std::this_thread::sleep_for(5ms);

	m_context.peval(js::Script{"timer.restart();"});

	std::this_thread::sleep_for(6ms);

	m_context.peval(js::Script{"result = timer.elapsed()"});
	assertRange(m_context.getGlobal<int>("result"), 16);
}

TEST_F(TestElapsedTimer, doublecheck)
{
	m_context.peval(js::Script{"timer = new Irccd.ElapsedTimer();"});

	std::this_thread::sleep_for(50ms);

	m_context.peval(js::Script{"result = timer.elapsed()"});

	std::this_thread::sleep_for(50ms);

	m_context.peval(js::Script{"result = timer.elapsed()"});
	assertRange(m_context.getGlobal<int>("result"), 100);
}

int main(int argc, char **argv)
{
	testing::InitGoogleTest(&argc, argv);

	return RUN_ALL_TESTS();
}
